/*
    tsync - a utility to sync remote encrypted copies of a tstore
    Copyright (C) 2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: tsync-server.c,v 1.13 2005/09/10 08:01:35 kasperd Exp $
*/

#include "tsync.h"

int efd;
off_t truncatelen=0;
uint64_t record_len=0;
uint32_t header[16];
uint64_t read_left=0;
#define HEADER ((uint8_t*)(header))

void errstop(const char *s)
{
  perror(s);
  exit(EXIT_FAILURE);
}

#define EXPECTED_SIZE (64+(ntoh64(header+14)<<3))

ssize_t myread(int fd, void *buf, size_t count)
{
  ssize_t r,total=0;
  memset(buf,0,count);
  while (count) {
    r=read(fd,buf,count);
    if (r<1) return total?total:r;
    total+=r;
    buf=(((char*)buf)+r);
    count-=r;
  }
  return total;
}

int verify_size()
{
  struct stat stbuf;
  if (fstat(efd,&stbuf)) return -1;
  return (stbuf.st_size<EXPECTED_SIZE);
}

int read_header()
{
  switch (myread(efd,HEADER,64)) {
  case 64: return (HEADER[7]||HEADER[51])?1:verify_size();
  case -1: return -1;
  default: return 1;
  }
}

int verify_cookie()
{
  uint8_t buf[43];
  switch (myread(0,buf,43)) {
  case 43: return (strncmp(HEADER+8,buf,43)!=0);
  case -1: return -1;
  default: return 1;
  }
}

#define must_be_locked() do { if(!locked) {	\
  fprintf(stderr,"%s: not locked\n",argv[1]);	\
  return EXIT_FAILURE;				\
 } } while (0)

#define must_be_unlocked() do { if (locked) {		\
  fprintf(stderr,"%s: already locked\n",argv[1]);	\
  return EXIT_FAILURE;					\
} } while (0)

void rollback()
{
  if (truncatelen) ftruncate(efd,truncatelen);
}

int lock_file_header()
{
  uint8_t newheader[64];
  int r;
  /* FIXME: Do locking here */
  struct flock mylock={
    .l_type=F_WRLCK,
    .l_whence=SEEK_SET,
    .l_start=52,
    .l_len=12,
  };
  if(fcntl(efd,F_SETLK,&mylock)) errstop("fcntl(F_SETLK)");

  switch(pread(efd,newheader,64,0)) {
  case -1: return -1;
  case 64: break;
  default: return 1;
  }
  if (memcmp(HEADER,newheader,64)) return 1;
  r=verify_size();
  if (r) return r;
  truncatelen=EXPECTED_SIZE;
  rollback();
  atexit(rollback);
  lseek(efd,truncatelen,SEEK_SET);
  return 0;
}

void select_command()
{
  uint32_t total_records=ntohl(*(uint32_t*)(HEADER+52));
  uint32_t selected_record;
  uint64_t start_of_record,end_of_record;
  uint32_t i;
  switch(myread(0,&selected_record,4)){
  case -1:
    errstop("read");
  case 4:
    break;
  default:
    fprintf(stderr,"Short read on stdin (select)\n");
    exit(EXIT_FAILURE);
  }
  selected_record=ntohl(selected_record);
  if (selected_record>=total_records) {
    fprintf(stderr,"No such record\n");
    exit(EXIT_FAILURE);
  }
  start_of_record=EXPECTED_SIZE;
  for (i=total_records;i>selected_record;--i) {
    uint32_t buf[2];
    end_of_record=start_of_record-8;
    switch(pread(efd,buf,8,end_of_record)){
    case -1:
      errstop("pread");
    case 8:
      break;
    default:
      fprintf(stderr,"Short pread on storage\n");
      exit(EXIT_FAILURE);
    }
    read_left=ntoh64(buf);
    start_of_record=end_of_record-((read_left+7)&~(uint64_t)7);
  }
  lseek(efd,start_of_record,SEEK_SET);
}

void read_command()
{
  uint32_t s,r;
  switch (myread(0,&s,4)) {
  case -1:
    errstop("read");
  case 4:
    break;
  default:
    fprintf(stderr,"Short read on stdin (read size)\n");
    exit(EXIT_FAILURE);
  }
  s=ntohl(s);
  if (s>read_left) s=read_left;
  r=htonl(s);
  switch(write(1,&r,4)) {
  case -1:
    errstop("write");
  case 4:
    break;
  default:
    fprintf(stderr,"Short write on stdout\n");
    exit(EXIT_FAILURE);
  }
  read_left-=s;
  while(s) {
    uint8_t buf[4096];
    unsigned max_read=sizeof(buf);
    int r,w;
    if (max_read>s) max_read=s;
    r=read(efd,buf,max_read);
    switch(r) {
    case -1:
      errstop("read");
    case 0:
      fprintf(stderr,"Short read on storage\n");
      exit(EXIT_FAILURE);
    }
    w=write(1,buf,r);
    if (w!=r) {
      if (w==-1) errstop("write");
      fprintf(stderr,"Short write on stdout\n");
      exit(EXIT_FAILURE);
    }
    s-=r;
  }
}

void write_command()
{
  uint32_t s;
  switch (myread(0,&s,4)) {
  case -1:
    errstop("read");
  case 4:
    break;
  default:
    fprintf(stderr,"Short read on stdin (write size)\n");
    exit(EXIT_FAILURE);
  }
  s=ntohl(s);
  record_len+=s;
  //fprintf(stderr,"%d\n",s);
  while(s) {
    uint8_t buf[4096];
    unsigned max_read=sizeof(buf);
    int r,w;
    if (max_read>s) max_read=s;
    r=read(0,buf,max_read);
    switch(r) {
    case -1:
      errstop("read");
    case 0:
      fprintf(stderr,"Short read on stdin (write data)\n");
      exit(EXIT_FAILURE);
    }
    w=write(efd,buf,r);
    if (w!=r) {
      if (w==-1) errstop("write");
      fprintf(stderr,"Short write on storage\n");
      exit(EXIT_FAILURE);
    }
    s-=r;
  }
}

void commit_command()
{
  int w,padlen=(8-(record_len&7))&7;
  uint64_t padding=0;
  uint32_t header_fields[3];
  w=write(efd,&padding,padlen);
  if (w!=padlen) {
    if (w==-1) errstop("write");
    fprintf(stderr,"Short write on storage\n");
    exit(EXIT_FAILURE);
  }
  switch(write(efd,hton64(record_len),8)) {
  case -1:
    errstop("write");
  case 8:
    break;
  default:
    fprintf(stderr,"Short write on storage\n");
    exit(EXIT_FAILURE);
  }
  header_fields[0]=htonl(ntohl(*(uint32_t*)(HEADER+52))+1);
  memcpy(header_fields+1,hton64((truncatelen+record_len-49)>>3),8);
  truncatelen=0;
  switch(pwrite(efd,header_fields,12,52)) {
  case -1:
    errstop("pwrite");
  case 12:
    break;
  default:
    fprintf(stderr,"Short pwrite on storage\n");
    exit(EXIT_FAILURE);
  }
  if (fsync(efd)) errstop("fsync");
  if (close(efd)) errstop("close");
  write(1,"Committed",9);
  exit(0);
}

int main(int argc, char ** argv)
{
  int locked=0;
  uint8_t command;

  if (argc!=2) {
    fprintf(stderr,"Usage: %s <encryptedstorage>\n",argv[0]);
    return EXIT_FAILURE;
  }

  efd=open(argv[1],O_RDWR);
  setgid(getgid()); /* Drop group priveleges? */

  /* Do you consider the error message an information leakage?
   * I'm not THAT paranoid. So what if an unpriveledged user
   * can find out if a named file exist and if it is a valid
   * encrypted storage? After all the storage is encrypted.
   * And the file name is going to be visible on the command
   * line anyway.
   *
   * If the caller cannot provide a valid cookie, he cannot do
   * anything with the storage anyway. And the error message
   * leaks no information about the cookie. The program is
   * intended to be set-group-id, and the group should have no
   * special priveleges beyond accessing the encrypted stores.
   */
  switch(read_header()){
  case -1:
    errstop(argv[1]);
  case 1:
    fprintf(stderr,"%s: bad storage format\n",argv[1]);
    return EXIT_FAILURE;
  }

  switch(verify_cookie()){
  case -1:
    errstop(argv[0]);
  case 1:
    fprintf(stderr,"%s: bad cookie\n",argv[1]);
    return EXIT_FAILURE;
  }

  if (write(1,HEADER+52,4)!=4) errstop("write");

  while (1) {
    switch(read(0,&command,1)) {
    case 0: return 0;
    case 1: break;
    default: errstop(argv[0]);
    }

    switch(command) {
    case 'S':
      must_be_unlocked();
      select_command();
      /*
      fprintf(stderr,"%s: select not implemented\n",argv[0]);
      return EXIT_FAILURE;
      */
      break;
    case 'R':
      must_be_unlocked();
      read_command();
      /*
      fprintf(stderr,"%s: read not implemented\n",argv[0]);
      return EXIT_FAILURE;
      */
      break;
    case 'L':
      must_be_unlocked();
      switch (lock_file_header()) {
      case -1:
	errstop(argv[1]);
      case 1:
	fprintf(stderr,"%s: was modified since opening\n",argv[1]);
	return EXIT_FAILURE;
      }
      locked=1;
      break;
    case 'W':
      must_be_locked();
      /*
      fprintf(stderr,"%s: write not implemented\n",argv[0]);
      return EXIT_FAILURE;
      */
      write_command();
      break;
    case 'C':
      must_be_locked();
      /*
      fprintf(stderr,"%s: commit not implemented\n",argv[0]);
      return EXIT_FAILURE;
      */
      commit_command();
      break;
    default:
      fprintf(stderr,"%s: unknown command %d\n",argv[0],command);
      return EXIT_FAILURE;
    }
  }
}
